/* makeDNS.c -- generate DNS RRs from /etc/hosts
 *
 * Copyright (c) 1991-1997 Department of Computer Science, University of
 * Edinburgh.
 * Copyright (c) 1998-2002 Division of Informatics, University of Edinburgh.
 *
 * Non-commercial use and redistribution is permitted provided this
 * notice remains intact and any modifications are flagged.
 * This software is offered "as is", with no warranty or support whatsoever.
 *
 * Please send comments to gdmr@dcs.ed.ac.uk.
 */

/*
 * $Id: makeDNS.c,v 1.17 2002/06/17 12:43:04 gdmr Exp $
 */

/* Generate a set of DNS RR files from /etc/hosts and some configuration
 * files, under a reasonable set of assumptions.  The program expects to
 * be running in the target directory.  It reads the file "Configure" in
 * that directory.  The file consists of a sequence of "wire" records, one
 * per line, each one listing the network number, netmask, RR file, header
 * file and full domain name for all hosts on that wire.  Alternatively,
 * a "domain" record consists of 0.0.0.0 as network number, domain, RR file
 * and header file.
 *
 * The /etc/hosts file is read in, and hosts are added to a 2-way linked list,
 * sorted alphabetically by host name (ordering in the DNS doesn't matter,
 * and this sorting allows the program's assumptions to be implemented).
 * Primary names are held with their corresponding IP address, while aliases
 * have the name of the corresponding primary.
 *
 * Lines in the hosts file beginning "#verbatim" are treated as follows:
 * white-space is skipped; then the first word is taken as the name of a
 * zone; more white-space is skipped; then the remainder of the line is saved
 * on a linked list, to be written later to the corresponding zone file.
 * Note that the RR must include a name, as the order in the output file
 * is not guaranteed.
 *
 * Once the hosts file has been read in, the output files are opened and
 * the constant header information is written (with the sequence number
 * properly updated).  The hosts list is then scanned in ascending order
 * and the corresponding entries written according to the following
 * heuristics: if a host name has as its leading component one of the
 * preceding hosts then it is assumed to be a second interface on a
 * multi-homed host, and an additional A record is written; and the domain
 * of a host is taken to be that of its first A record, as defined in the
 * corresponding "wire" entry.  Finally the program exits with a zero status.
 *
 * Should any errors occur, the program will exit with a non-zero status.
 */

static char *RCS_revision = "$Revision: 1.17 $";

#define DEFAULT_HOSTS "/etc/nis/src/hosts"
#define DEFAULT_CONFIG "Configure"
#define LOG "Log.DNS"
#define MAXWIRE 127
#define MAXDOMAIN 5
#define LINEBUFFER 1024
#define DOT_LIMIT 3

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>

extern char *optarg;

typedef struct _verbatim {
	struct _verbatim *next;
	char data[LINEBUFFER];
} verbatim;

typedef struct _domain {
	char *domainname;
	char *RRfile;
	char *headerfile;
	char *otherZonefile;
	char *MXfile;
	FILE *file;
	FILE *other;
	char *MXdata;
	verbatim *verbData;
} domain;

typedef struct _wire {
	long number;
	long netmask;
	char *RRfile;
	char *headerfile;
	char *domainname;
	FILE *file;
	domain *d;
	char *LOCfile;
	char *LOCdata;
} wire;

typedef struct _host {
	struct _host *forward;
	struct _host *backward;
	long IP;
	char *name;
	struct _host *primary;
	wire *w;
} host;

static host *hostlist;
static wire wireTable[MAXWIRE];
static domain domainTable[MAXDOMAIN];
static int domains, wires;

static char *hostsFile, *configFile;

static void readconfig();
static void readhosts();
static void printhosts();
static void openfiles();
static void closefiles();

main(int argc, char **argv)
{	int c;
	int err = 0;
	
	while ((c = getopt(argc, argv, "c:h:")) != EOF) {
		switch(c) {
		case 'c':
			configFile = optarg;
			break;

		case 'h':
			hostsFile = optarg;
			break;

		default:
			err++;
			break;
		}
	}

	if (err) {
		(void) fprintf(stderr,
				"Usage: makeDNS -c <config-file> -h <hostname-file>\n");
		exit(5);
	}

	if (!configFile) configFile = DEFAULT_CONFIG;
	if (!hostsFile) hostsFile = DEFAULT_HOSTS;

	(void) printf("makeDNS: config file %s, hosts source file %s\n",
			configFile, hostsFile);

	readconfig();
	readhosts();
	openfiles();
	printhosts();
	closefiles();
	return 0;
}

static int countdots(name)
char *name;
{	int count = 0;

	if (!name) return 0;

	while (*name) {
		if (*name == '.') count++;
		name++;
	}

	return count;
}

static void addhost(h)
host *h;
{	host *x;
	int dotted;
	if (hostlist == NULL) {
		/* List empty, easy case */
		h->forward = NULL;
		h->backward = NULL;
		hostlist = h;
		return;
	}
	x = hostlist;
	/* Skip to insertion point */
	while (strcmp(x->name, h->name) < 0) {
		if (x->forward) {
			/* More, skip */
			x = x->forward;
		}
		else {
			/* Add at the tail of the list */
			h->forward = NULL;
			h->backward = x;
			x->forward = h;
			return;
		}
	}
	/* Sanity check: duplicate?  Only do this for our own hosts. */
	dotted = (countdots(h->name) >= DOT_LIMIT);
	if (!dotted && !strcmp(x->name, h->name)) {
		(void) fprintf(stderr, "** %s duplicate!\n", h->name);
		return;
	}
	/* Insert before x */
	h->forward = x;
	h->backward = x->backward;
	if (x->backward) x->backward->forward = h;	/* Middle */
	else hostlist = h;				/* Head */
	x->backward = h;
}

static long IPconvert(ch)
char **ch;
{	long N1 = -1, N2 = -1, N3 = -1, N4 = -1;

	while (**ch && isspace(**ch)) (*ch)++;

	/* Convert one component of a dotted-quad to its binary
	 * representation.  Assume there may be three decimal digits,
	 * do nothing if the first character isn't one, and stop
	 * when the second or third aren't.  And complain if the result
	 * is too big, since this is supposed to be an octet value.
	 */
#define ConvertOne(X) \
	X = -1; \
	if (**ch >= '0' && **ch <= '9') {X = **ch - '0'; (*ch)++; \
	if (**ch >= '0' && **ch <= '9') {X = (10 * X) + **ch - '0'; (*ch)++; \
	if (**ch >= '0' && **ch <= '9') {X = (10 * X) + **ch - '0'; (*ch)++; \
	} } } \
	if (X > 255) X = -1;

	/* Apply the above to all four components of the dotted quad.
	 * If one of them doesn't convert properly or there isn't a '.'
	 * between the fields or white space at the end then fail. */
	ConvertOne(N1);  if (N1 < 0 || **ch != '.') return -1;  (*ch)++;
	ConvertOne(N2);  if (N2 < 0 || **ch != '.') return -1;  (*ch)++;
	ConvertOne(N3);  if (N3 < 0 || **ch != '.') return -1;  (*ch)++;
	ConvertOne(N4);  if (N4 < 0 || !isspace(**ch)) return -1;  (*ch)++;

#undef ConvertOne

	return (N1 << 24) | (N2 << 16) | (N3 << 8) | N4;
}

static char *extractname(ch)
char **ch;
{	char *it;
	char *name;
	int n = 0;
	while (**ch && isspace(**ch)) (*ch)++;
	if (**ch == '\0' || **ch == '#') return NULL;
	it = *ch;
	while (**ch && !isspace(**ch)) {
		(*ch)++;
		n++;
	}
	if (**ch) {
		**ch = '\0';
		(*ch)++;
	}
	name = (char *) malloc(n + 1);
	strcpy(name, it);
	return name;
}

static wire *findWire(IP)
long IP;
{	int n;
	wire *w;
	if (IP == 0) return NULL;
	for (n = 0, w = wireTable; n < wires; n++, w++) {
		if ((IP & w->netmask) == w->number) {
			return w;
		}
	}
	return NULL;
}

static domain *findDomain(name)
char *name;
{	int n;
	domain *d;
	for (n = 0, d = domainTable; n < domains; n++, d++) {
		if (!strcmp(name, d->domainname)) {
			return d;
		}
	}
	return NULL;
}

static void doComment(line)
char *line;
{	verbatim *v;
	domain *d;
	char *name;
	char *data;

	if (!strncmp(line, "#verbatim", 9)) {
		line += 9;
		while (*line != '\0') {
			if (*line != ' ' && *line != '\t') break;
			line++;
		}
		if (*line == '\0' || *line == '\n') {
			(void) fprintf(stderr, "** Verbatim format error\n");
			return;
		}
		name = line;
		while (*line != ' ' && *line != '\t') {
			if (*line == '\0' || *line == '\n') {
				(void) fprintf(stderr,
						"** Verbatim format error\n");
				return;
			}
			line++;
		}
		*line++ = '\0';
		data = line;
		while (*data != '\0') {
			if (*data != ' ' && *data != '\t') break;
			data++;
		}
		if (*data == '\0' || *data == '\n') {
			(void) fprintf(stderr, "** Verbatim format error\n");
			return;
		}
		d = findDomain(name);
		if (!d) {
			fprintf(stderr,
				"** Warning: can't find #verbatim name \"%s\"\n",
				name);
			return;
		}
		v = (verbatim *) malloc(sizeof(verbatim));
		strcpy(v->data, data);
		v->next = d->verbData;
		d->verbData = v;
	}
}

static int hostOK(h)
char *h;
{	char *ch;
	if (h && *h) {
		ch = h;
		if (!isalnum(*ch) && *ch != '.' && *ch != '-') {
			fprintf(stderr,
				"** Illegal character in <%s>\n", h);
			return 0;
		}
		ch++;
		while (*ch) {
			if (!isalnum(*ch) && *ch != '.' && *ch != '-') {
				fprintf(stderr,
					"** Illegal character in <%s>\n", h);
				return 0;
			}
			ch++;
		}
		return 1;
	} else {
		fprintf(stderr, "hostOK(): null name or pointer\n");
		return 0;
	}
}

static void readhosts()
{	FILE *hosts;
	host *h, *ph;
	char *ch, *end;
	char buffer[LINEBUFFER];
	long IP;
	char *primary;
	char *alias;
	wire *w;
	int errors = 0;

	if ((hosts = fopen(hostsFile, "r")) == NULL) {
		(void) fprintf(stderr, "** makeDNS: failed to open %s\n",
				hostsFile);
		exit(1);
	}

	for (;;) {
		/* Grab the next line and chop off the trailing newline */
		if ((ch = fgets(buffer, LINEBUFFER, hosts)) == NULL) break;
		end = ch + strlen(ch) - 1;
		if (*end == '\n') *end = '\0';

#ifdef DEBUG
		(void) fprintf(stderr, "Processing <%s>\n", buffer);
#endif

		/* Leading junk */
		while (*ch && isspace(*ch)) ch++;

		/* Ignore null lines */
		if (*ch == '\0') continue;

		/* Process comments.  Note they may be "verbatim" */
		if (*ch == '#') {
			doComment(ch);
			continue;
		}

		/* Extract the IP address and find its wire */
		IP = IPconvert(&ch);
		if (IP == -1) {
			(void) fprintf(stderr,
					"** Bogus IP addr in <%s>\n",
					buffer);
			continue;
		}
		if ((w = findWire(IP)) == NULL) {
#ifndef DONT_WARN
			if (IP && (IP != 0x7f000001)) {
				while (*ch && isspace(*ch)) ch++;
				(void) fprintf(stderr,
						"** Unknown subnet: %s", ch);
			}
#endif
			continue;
		}

		/* Extract the hostname */
		primary = extractname(&ch);
		if (!primary) {
			(void) fprintf(stderr,
				"** Bogus hostname in <%s>\n",
				buffer);
			continue;
		}
		ph = (host *) malloc(sizeof(host));
		ph->IP = IP;
		ph->w = w;
		ph->name = primary;
		ph->primary = NULL;
		if (hostOK(primary)) addhost(ph);
		else errors++;

		/* Any aliases or wire-names? */
		while (*ch != '\0') {
			alias = extractname(&ch);
			if (alias == NULL) break;
			h = (host *) malloc(sizeof(host));
			h->IP = 0;
			h->w = w;
			h->name = alias;
			h->primary = ph;
			if (hostOK(alias)) addhost(h);
			else errors++;
		}
	}
	(void) fclose(hosts);
	if (errors) {
		(void) fprintf(stderr, "** Hosts file %s contained errors\n",
				hostsFile);
		exit(2);
	}
}

static host *findPrincipal(like)
host *like;
{	host *h;
	int l, n;

	h = like->backward;
	l = strlen(like->name);
	while (h) {
		if (*(h->name) != *(like->name)) return NULL;
		n = strlen(h->name);
		if ((n < l) && /* !isalnum(like->name[n]) && */
				(like->name[n] == '-') &&
				!strncmp(h->name, like->name, n)) {
			return h;
		}
		h = h->backward;
	}
	return NULL;
}

#define  printIP(file, IP) \
	(void) fprintf(file, "%d.%d.%d.%d", \
		(IP >> 24) & 255, (IP >> 16) & 255, (IP >> 8) & 255, IP & 255)

/* printhosts has four cases:
 *   1	IP address, no principal entry -- write the name as it stands, plus
 *	the reverse name, again as it stands.
 *   2	IP address plus principal entry -- write out the address as for the
 *	principal entry, plus the current name as another A RR.  Write the
 *	reverse entry as pointing to the principal name.
 *   3	alias, no principal entry -- write a CNAME RR pointing to the primary.
 *   4	alias plus principal entry -- write an A RR with the primary's address.
 *
 * For example, suppose the /etc/hosts file has the following:
 *	129.215.64.48  baleshare baleshare-b agfahost
 *	29.215.160.155 baleshare-gw baleshare-a
 * Then baleshare is case 1, baleshare-gw is case 2, baleshare-a and
 * baleshare-b are case 4, and agfahost is case 3.  Confused?  BTW, changing
 * baleshare-gw to fred, say, would break the algorithm -- the main name of the
 * second interface better be related to the main name of the first!
 */

void printhosts()
{	host *h, *m, *x;
	wire *w;
	FILE *f;
	int dotted;

	h = hostlist;
	while (h) {
		m = findPrincipal(h);
		if (h->IP) {
			if (m) {
				/* Case 2 */
#ifdef DEBUG
				printIP(stderr, h->IP);
				(void) fprintf(stderr, " -- case 2 <%s>\n",
					h->name);
#endif
				/* A RR for name */
				(void) fprintf(m->w->d->file,
					"%-23s IN\tA\t", h->name);
				printIP(m->w->d->file, h->IP);
				(void) putc('\n', m->w->d->file);
				if (m->w->d->MXdata) {
					(void) fputs(m->w->d->MXdata,
						m->w->d->file);
				}
				/* A RR for principal's name */
				(void) fprintf(m->w->d->file,
					"%-23s IN\tA\t", m->name);
				printIP(m->w->d->file, h->IP);
				(void) putc('\n', m->w->d->file);
				/* back-pointer -> principal */
				(void) fprintf(h->w->file,
					"%d\tIN\tPTR\t%s.%s.\n",
					h->IP & (~h->w->netmask),
					m->name, m->w->d->domainname);
				if (m->w->d->other) {
					(void) fprintf(m->w->d->other,
						"%-23s IN\tCNAME\t%s.%s.\n",
						h->name, m->name,
						m->w->d->domainname);
				}
			}
			else {
				/* Case 1 */
#ifdef DEBUG
				printIP(stderr, h->IP);
				(void) fprintf(stderr, " -- case 1 <%s>\n",
					h->name);
#endif
				dotted = (countdots(h->name) >= DOT_LIMIT);
				if (!dotted) {
					(void) fprintf(h->w->d->file,
						"%-23s IN\tA\t", h->name);
					printIP(h->w->d->file, h->IP);
					(void) putc('\n', h->w->d->file);
					if (h->w->d->MXdata) {
						(void) fputs(h->w->d->MXdata,
							h->w->d->file);
					}
					if (h->w->LOCdata) {
						(void) fputs(h->w->LOCdata,
							h->w->d->file);
					}
					if (h->w->d->other) {
						(void) fprintf(h->w->d->other,
							"%-23s IN\tCNAME\t%s.%s.\n",
							h->name, h->name,
							h->w->d->domainname);
					}
				}
				(void) fprintf(h->w->file,
					dotted ? "%d\tIN\tPTR\t%s\n"
						: "%d\tIN\tPTR\t%s.%s.\n",
					h->IP & (~h->w->netmask),
					h->name, h->w->d->domainname);
			}
		}
		else {
			if (m) {
				/* Case 4 */
#ifdef DEBUG
				(void) fprintf(stderr, "Case 4 <%s>\n",
					h->name);
#endif
				(void) fprintf(m->w->d->file,
					"%-23s IN\tA\t", h->name);
				printIP(m->w->d->file, h->primary->IP);
				(void) putc('\n', m->w->d->file);
				if (m->w->d->MXdata) {
					(void) fputs(m->w->d->MXdata,
						m->w->d->file);
				}
				if (m->w->d->other) {
					(void) fprintf(m->w->d->other,
						"%-23s IN\tCNAME\t%s.%s.\n",
						h->name, m->name,
						m->w->d->domainname);
				}
			}
			else {
				/* Case 3 */
#ifdef DEBUG
				(void) fprintf(stderr, "Case 3 <%s>\n",
					h->name);
#endif
				x = findPrincipal(h->primary);
				if (!x) x = h->primary;
				(void) fprintf(x->w->d->file,
					"%-23s IN\tCNAME\t%s.%s.\n",
					h->name, x->name,
					x->w->d->domainname);
				if (x->w->d->other) {
					(void) fprintf(x->w->d->other,
						"%-23s IN\tCNAME\t%s.%s.\n",
						h->name, x->name,
						x->w->d->domainname);
				}
			}
		}
		h = h->forward;
	}
}

static void readconfig()
{	FILE *config;
	long IP;
	char *ch, *end;
	char buffer[LINEBUFFER];
	char next[LINEBUFFER];
	wire *w;
	domain *d;

	if ((config = fopen(configFile, "r")) == NULL) {
		(void) fprintf(stderr, "** makeDNS: failed to open %s\n",
				configFile);
		exit(1);
	}
	for (;;) {
		if (fgets(buffer, LINEBUFFER, config) == NULL) break;
		end = buffer + strlen(buffer) - 1;
		if (*end == '\n') *end = '\0';

		/* If the last character in the line was a '\' then
		 * chop it off, grab the next line, glue the two
		 * together and try again.  Only do this, though, if
		 * the line wasn't completely blank in the first place!
		 */
		if (*buffer) {
			end--;
			while (*end == '\\') {
				*end = '\0';  /* Zap that '\' */
				if (fgets(next, LINEBUFFER, config) == NULL) break;
				end = next + strlen(next) - 1;
				if (*end == '\n') *end = '\0';
				strcat(buffer, next);
				end = buffer + strlen(buffer) - 1;
			}
		}

#ifdef DEBUG
		(void) fprintf(stderr, "Processing <%s>\n", buffer);
#endif

		/* Leading junk */
		ch = buffer;
		while (*ch && isspace(*ch)) ch++;
		if (*ch == '\0' || *ch == '#') continue;

		/* First field should be a network number */
		IP = IPconvert(&ch);
		if (IP == -1) {
			(void) fprintf(stderr,
					"** Bogus address in config file\n");
			(void) fputs(buffer, stderr);
			(void) fputc('\n', stderr);
			exit(2);
		}
		else if (IP) {
			/* Wire record */
			w = wireTable + wires++;
			w->number = IP;
			w->netmask = IPconvert(&ch);
			if (w->netmask == -1) {
				(void) fprintf(stderr,
					"** Bogus netmask in config file\n");
				(void) fputs(buffer, stderr);
				(void) fputc('\n', stderr);
				exit(2);
			}
			w->RRfile = extractname(&ch);
			w->headerfile = extractname(&ch);
			w->domainname = extractname(&ch);
			w->file = stderr;  /* meantime */
			w->d = findDomain(w->domainname);
			if (!w->d) {
				(void) fprintf(stderr,
					"** Undefined domain %s:\n",
					w->domainname);
				(void) fputs(buffer, stderr);
				(void) fputc('\n', stderr);
				exit(2);
			}
			w->LOCfile = extractname(&ch);
			w->LOCdata = NULL;
#ifdef DEBUG
	fprintf(stderr, "Wire: %x %x %s %s %s %s\n", w->number, w->netmask,
		w->RRfile, w->headerfile, w->domainname, w->LOCfile);
#endif
		}
		else {
			/* Domain record */
			d = domainTable + domains++;
			d->domainname = extractname(&ch);
			d->RRfile = extractname(&ch);
			d->headerfile = extractname(&ch);
			d->otherZonefile = extractname(&ch);
			d->MXfile = extractname(&ch);
			d->file = stderr;  /* meantime */
			d->other = NULL;
			d->MXdata = NULL;
			d->verbData = NULL;
		}
	}
	(void) fclose(config);
}

static void writeheader(f, h, MXdata, verbData)
FILE *f;
char *h;
char *MXdata;
verbatim *verbData;
{	static struct passwd *pwd;
	static char *pwname;
	time_t tt;
	static struct tm *ttm;
	static char *t;
	static char *ln;
	static char stamp[20];
	FILE *hf, *l;
	char buffer[LINEBUFFER];
	int doingVerbatim = (int) verbData;

	if (!ln) ln = getenv("LOGNAME");
	if (!ln) {
		fprintf(stderr, "** LOGNAME undefined\n");
		ln = "**unknown**";
	}

	if (!pwd) {
		pwd = getpwuid(getuid());
		endpwent();
		if (pwd) pwname = pwd->pw_name;
		else pwname = "UNKNOWN";
		tt = time(NULL);
		ttm = localtime(&tt);
		(void) sprintf(stamp, "%d%3.3d%3.3d",
			1900 + ttm->tm_year, ttm->tm_yday,
			((60 * ttm->tm_hour) + ttm->tm_min) >> 1);
		t = asctime(ttm);
		t[strlen(t) - 1] = '\0';  /* Zap the trailing \n */
		if (l = fopen(LOG, "a")) {
			fprintf(l, "Zone generated by %s (%s) on %s using makeDNS %s\n",
					pwname, ln, t, RCS_revision);
			fclose(l);
		}
	}
	(void) fprintf(f, ";; Zone generated by %s (%s) on %s\n",
			pwd->pw_name, ln, t);
	(void) fprintf(f, ";; using makeDNS %s\n\n",RCS_revision);
	if ((hf = fopen(h, "r")) == NULL) {
		(void) fprintf(stderr, "** Can't open header file %s\n", h);
		exit(4);
	}
	(void) fprintf(f, ";; Including %s\n", h);
	for (;;) {
		if (fgets(buffer, LINEBUFFER, hf) == NULL) break;
		(void) fprintf(f, buffer, stamp);
	}
	(void) fclose(hf);
	(void) fprintf(f, ";; End of %s\n", h);
	(void) fprintf(f, "\ngenerated\tIN\tTXT\t\"Zone generated by %s (%s) on %s using makeDNS %s\"\n\n",
		pwd->pw_name, ln, t, RCS_revision);
	if (MXdata) {
		(void) fputc('@', f);
		(void) fputs(MXdata, f);
		(void) fputc('\n', f);
	}
	if (doingVerbatim) (void) fputs(";; Verbatim data\n\n", f);
	while (verbData) {
		(void) fputs(verbData->data, f);
		(void) fputc('\n', f);
		verbData = verbData->next;
	}
	if (doingVerbatim) (void) fputs("\n;; End of verbatim data\n\n", f);
}

static void openfiles()
{	int i;
	int MXfd;
	int MXsize;
	int LOCfd;
	int LOCsize;
	wire *w;
	domain *d;

	for (i = 0, d = domainTable; i < domains; i++, d++) {
		if ((d->file = fopen(d->RRfile, "w")) == NULL) {
			(void) fprintf(stderr, \
				"** Can't open RRfile%s\n", d->RRfile);
			exit(3);
		}
		if (d->otherZonefile) {
			if ((d->other = fopen(d->otherZonefile, "w")) == NULL) {
				(void) fprintf(stderr, \
					"** Can't open otherZonefile %s\n",
					d->otherZonefile);
				exit(3);
			}
		}
		if (d->MXfile) {
			if ((MXfd = open(d->MXfile, O_RDONLY, 0)) < 0) {
				(void) fprintf(stderr, \
					"** Can't open MXfile %s\n",
					d->MXfile);
				exit(3);
			}
			d->MXdata = (char *) malloc(2048);
			if ((MXsize = read(MXfd, d->MXdata, 2048)) < 0) {
				(void) fprintf(stderr,
						"** Couldn't read from %s\n",
						d->MXfile);
				exit(3);
			}
			(void) close(MXfd);
			*(d->MXdata + MXsize) = '\0';
		}
		writeheader(d->file, d->headerfile, d->MXdata, d->verbData);
		if (d->otherZonefile)
			writeheader(d->other, d->headerfile, d->MXdata, NULL);
	}
	for (i = 0, w = wireTable; i < wires; i++, w++) {
		if ((w->file = fopen(w->RRfile, "w")) == NULL) {
			(void) fprintf(stderr, \
				"** Can't open RRfile %s\n", w->RRfile);
			exit(3);
		}
		if (w->LOCfile) {
			if ((LOCfd = open(w->LOCfile, O_RDONLY, 0)) < 0) {
				(void) fprintf(stderr, \
					"** Can't open LOCfile %s\n",
					w->LOCfile);
				exit(3);
			}
			w->LOCdata = (char *) malloc(2048);
			if ((LOCsize = read(LOCfd, w->LOCdata, 2048)) < 0) {
				(void) fprintf(stderr, "Couldn't read from %s\n",
						w->LOCfile);
				exit(3);
			}
			(void) close(LOCfd);
			*(w->LOCdata + LOCsize) = '\0';
		}
		writeheader(w->file, w->headerfile, (char *) 0, NULL);
	}
}

static void closefiles()
{	int i;
	domain *d;
	wire *w;

	for (i = 0, d = domainTable; i < domains; i++, d++) {
		(void) fclose(d->file);
		if (d->other) (void) fclose(d->other);
	}
	for (i = 0, w = wireTable; i < wires; i++, w++) {
		(void) fclose(w->file);
	}
}

/*
 * $Log: makeDNS.c,v $
 * Revision 1.17  2002/06/17 12:43:04  gdmr
 * Add a little more debugging.
 *
 * Revision 1.16  2002/06/14 15:49:24  gdmr
 * Tidier output file headers.
 *
 * Revision 1.15  2002/06/14 15:46:34  gdmr
 * Sanity-check only our own names.
 * Config-file continuation lines.
 * Make some error messages clearer.
 *
 * Revision 1.14  1999/11/17 11:02:45  gdmr
 * Division copyright
 *
 * Revision 1.13  1999/11/15 14:45:51  gdmr
 * Check for allowed characters in hostnames
 *
 * Revision 1.12  1999/10/05 12:50:04  gdmr
 * <unistd.h> and optarg
 *
 * Revision 1.11  1999/04/07 14:32:16  gdmr
 * Add wire-specific LOC-file support
 * (an extra file spec on the end of the "wire" definitions).
 *
 * Revision 1.10  1998/09/08 09:25:31  gdmr
 * RCS_revision in more places
 *
 * Revision 1.9  1998/09/08 09:20:18  gdmr
 * Add revision number to "generated"
 *
 * Revision 1.8  1998/09/08 09:11:13  gdmr
 * Don't fail if LOGNAME isn't set.  Grumble instead.
 *
 * Revision 1.7  1998/08/19 15:25:14  gdmr
 * "generated..." now called "generated" instead of '@'
 *
 * Revision 1.6  1997/08/11  10:56:50  gdmr
 * Better Y2K provision
 *
 * Revision 1.5  1997/08/11  10:50:21  gdmr
 * Better error checking for bogus IP addresses
 *
 * Revision 1.4  1997/08/08  15:21:09  gdmr
 * More sanity checking in readhosts()
 *
 * Revision 1.3  1997/08/08  12:48:45  gdmr
 * More RCS stuff
 *
 * Revision 1.2  1997/08/08  12:44:58  gdmr
 * RCS stuff
 *
 *
 */
